require("socket") -- LuaSocket
require("ex") -- Extension Proposal

dofile("config.txt")

function string.split(txt,d)
	if not d then d = " " end
	local split = {}
	local splitpos = {}
	local lastPos = 1
	for i = 1, string.len(txt)+1 do
		if string.sub(txt,i,i+string.len(d)-1) == d or i == string.len(txt)+1 then
			local newSplit = string.sub(txt,lastPos,i-1)
			table.insert(split,newSplit)
			table.insert(splitpos, lastPos)
			lastPos = i + string.len(d)
		end
	end
	return split, splitpos
end

FTP = {}
FTP.Client = {}
FTP.Queue = {}

function FTP.Login(name,password)
	return true
end

function FTP.InitServer()
	local e
	FTP.Server, e = socket.tcp()

	if FTP.Server then
		local success, e = FTP.Server:bind("*", ftp_port)
		if success then
			local success, e = FTP.Server:listen()
			if success then
				FTP.Server:settimeout(0, "t")
				FTP.Server:settimeout(0, "b")

				local ip, port = FTP.Server:getsockname()
				print("[FTP] Initialized FTP server  "..ip..":"..port)
			else
				FTP.Server:close()
				FTP.Server = nil
			end
		else
			FTP.Server:close()
			FTP.Server = nil
		end
	end
end

function FTP.RemoveQueue(id, success)
	if FTP.Queue[id] then
		FTP.Queue[id].Socket:close()
		FTP.Queue[id].file:close()
		FTP.Queue[id] = nil
	end
end

function FTP.ProcessClient(id,line)
	local Client = FTP.Client[id]

	if line then
		if string.sub(line, 1, 4) == "USER" then
			local UserName = string.sub(line, 6)
			if UserName:len() > 0 then
				Client.User = UserName
				Client.Socket:send("331 Password required to access user account "..UserName..".\n")
			end
			return true
		elseif string.sub(line, 1, 4) == "PASS" then
			local Password = string.sub(line, 6)
			if Password:len() > 0 then
				if FTP.Login(Client.User, Password) then
					Client.Password = Password
					Client.Logged = true
					Client.Socket:send("230 Logged in.\n")
				else
					Client.Socket:send("530 Login authentication failed\n")
				end
			else
				Client.Socket:send("331 Password required to access user account "..UserName..".\n")
			end
			return true
		elseif string.sub(line, 1, 3) == "CWD" then
			if Client.Logged then
				local Folder = string.sub(line, 5)
				while string.sub(Folder, 1, 1) == "/" do
					Folder = string.sub(Folder, 2)
				end

				if Folder == ".." then
					local find
					for i = 1, #Client.Dir do
						if string.sub(Client.Dir, i, i) == "/" then
							find = i
						end
					end

					if find then
						Client.Dir = string.sub(Client.Dir, 1, find - 1) or ""
						Client.Socket:send('250 "/'..Client.Dir..'" is new working directory\n')
					else
						Client.Dir = ""
						Client.Socket:send('250 "/'..Client.Dir..'" is new working directory\n')
					end
				elseif Folder:len() > 0 then
					if line:sub(5, 5) == "/" or Client.Dir:len() <= 0 then
						Client.Dir = Folder
					else
						Client.Dir = Client.Dir.."/"..Folder
					end
					Client.Socket:send('250 "/'..Client.Dir..'" is new working directory\n')
				else
					Client.Dir = ftp_dir
					Client.Socket:send('250 "/'..Client.Dir..'" is new working directory\n')
				end
			else
				Client.Socket:send("530 You aren't logged in\n")
			end
			return true
		elseif string.sub(line, 1, 4) == "SYST" then
			Client.Socket:send("215 Lua Virtual Machine\n")
			return true
		elseif string.sub(line, 1, 4) == "TYPE" then
			if Client.Logged then
				local NewType = string.sub(line, 6)
				if NewType == "I" then
					Client.Socket:send("200 TYPE is now 8-bit binary\n")
				elseif NewType == "A" then
					Client.Socket:send("200 TYPE is now ASCII\n")
				end
			else
				Client.Socket:send("530 You aren't logged in\n")
			end
			return true
		elseif string.sub(line, 1, 4) == "LIST" then
			if Client.Logged then
				if not Client.LSTART then
					Client.LSTART = os.clock()
				end

				if Client:ConnectData() then
					Client.Socket:send("150 Opening ASCII mode data connection for file list.\n")
					local Dir
					if Client.Dir:len() > 0 then
						Dir = Client.Dir
					else
						Dir = os.currentdir()
					end

					if os.dirent(Dir) then
						for entry in os.dir(Dir) do
							if entry.type == "directory" then
								Client.DataSocket:send("drwxr--r--	1	"..Client.User.."	PC	"..id.."	"..entry.size.."	"..entry.name.."\n")
							elseif entry.type == "file" then
								Client.DataSocket:send("-rwxr--r--	1	"..Client.User.."	PC	"..id.."	"..entry.size.."	"..entry.name.."\n")
							end
						end
					end
					Client.Socket:send("226 Transfer complete.\n")
					Client:CloseData()
					Client.LSTART = nil
					return true
				elseif os.clock() - Client.LSTART >= 2 then
					Client.Socket:send("425 Can't open data connection\n")
					Client.LSTART = nil
					return true
				end
			else
				Client.Socket:send("530 You aren't logged in\n")
				return true
			end
		elseif string.sub(line, 1, 4) == "PASV" then
			if Client.Logged then
				Client:ClosePasv()
				local sock, e = socket.tcp()
				if sock then
					local success, e = sock:bind("*", 0)
					if success then
						success, e = sock:listen()
						if success then
							sock:settimeout(0, "b")
							sock:settimeout(0, "t")

							local ip, port = sock:getsockname()
							local ports = {}
							ports[1] = math.floor(port/256)
							ports[2] = port - ports[1]*256

							local ips = ftp_ip:split(".")

							Client.Socket:send("227 Entering Passive Mode ("..ips[1]..","..ips[2]..","..ips[3]..","..ips[4]..","..ports[1]..","..ports[2]..")\n")
							Client.PasvServer = sock
							Client.CONMODE = 1
						else
							sock:close()
						end
					else
						sock:close()
					end
				else
					sock:close()
				end
			else
				Client.Socket:send("530 You aren't logged in\n")
			end
			return true
		elseif string.sub(line, 1, 4) == "PORT" then
			if Client.Logged then
				Client:CloseConMode()
				local adr = {string.match(line, "PORT (%d+),(%d+),(%d+),(%d+),(%d+),(%d+)")}
				if #adr >= 6 then
					local ip = adr[1].."."..adr[2].."."..adr[3].."."..adr[4]
					local port = tonumber(adr[5])*256 + tonumber(adr[6])

					if ip == Client:ip() then
						Client.DataSocket = DataSocket
						Client.Socket:send("200 PORT command successful\n")
						Client.DIP = ip
						Client.DPORT = port
						Client.CONMODE = 2
					else
						Client.Socket:send("500 I won't open a connection to "..ip.." (only to "..Client:ip()..")\n")
					end
				else
					Client.Socket:send("500 I won't open a connection\n")
				end
			else
				Client.Socket:send("530 You aren't logged in\n")
			end
			return true
		elseif string.sub(line, 1, 3) == "MKD" then
			if Client.Logged then
				local Folder = string.sub(line, 5)
				local Dir
				while string.sub(Folder, 1, 1) == "/" do
					Folder = string.sub(Folder, 2)
				end

				if Client.Dir:len() > 0 then
					Dir = Client.Dir.."/"..Folder
				else
					Dir = Folder
				end

				local success, e = os.mkdir(Dir)
				if success then
					Client.Socket:send('257 "'..Folder..'" : The directory was successfully created\n')
				else
					local e = e or "File exists"
					Client.Socket:send("550 Can't create directory: "..e.."\n")
				end
			else
				Client.Socket:send("530 You aren't logged in\n")
			end
			return true
		elseif string.sub(line, 1, 3) == "RMD" then
			if Client.Logged then
				local Folder = string.sub(line, 5)
				local Dir
				while string.sub(Folder, 1, 1) == "/" do
					Folder = string.sub(Folder, 2)
				end

				if Client.Dir:len() > 0 then
					Dir = Client.Dir.."/"..Folder
				else
					Dir = Folder
				end

				local success, e = os.remove(Dir)
				if success then
					--250 The directory was successfully removed
					Client.Socket:send("250 The directory was successfully removed\n")
				else
					Client.Socket:send("550 Can't remove directory: "..e.."\n")
				end
			else
				Client.Socket:send("530 You aren't logged in\n")
			end
			return true
		elseif string.sub(line, 1, 4) == "PWD" then
			if Client.Logged then
				Client.Socket:send('257 "/'..Client.Dir..'" is the current working directory\n')
			else
				Client.Socket:send("530 You aren't logged in\n")
			end
			return true
		elseif string.sub(line, 1, 4) == "RNFR" then
			local File = string.sub(line, 6)
			local Dir
			if Client.Dir:len() > 0 then
				Dir = Client.Dir.."/"..File
			else
				Dir = File
			end

			if File:len() <= 0 then
				Client.Socket:send("550 No file name\n")
			elseif File:len() > 0 then
				if os.dirent(Dir) then
					Client.Socket:send("350 RNFR accepted - file exists, ready for destination\n")
					Client.RenameFrom = File
				else
					Client.Socket:send("550 Sorry, but that file doesn't exist\n")
				end
			end
			return true
		elseif string.sub(line, 1, 4) == "RNTO" then
			if not Client.RenameFrom or Client.RenameFrom:len() <= 0 then
				Client.Socket:send("503 Need RNFR before RNTO\n")
				return true
			end

			local File = string.sub(line, 6)
			local RenameDir, SourceDir
			if Client.Dir:len() > 0 then
				RenameDir = Client.Dir.."/"..File
				SourceDir = Client.Dir.."/"..Client.RenameFrom
			else
				RenameDir = File
				SourceDir = Client.RenameFrom
			end

			if File:len() <= 0 then
				Client.Socket:send("550 No file name\n")
			elseif File:len() > 0 then
				if os.dirent(RenameDir) then
					Client.Socket:send("550 RENAME Failed - destination file already exists\n")
				else
					Client.Socket:send("250 File successfully renamed\n")
					os.rename(SourceDir, RenameDir)
				end
			end
			return true
		elseif string.sub(line, 1, 4) == "DELE" then
			local File = string.sub(line, 6)
			local Dir
			while string.sub(File, 1, 1) == "/" do
				File = string.sub(File, 2)
			end

			if Client.Dir:len() > 0 then
				Dir = Client.Dir.."/"..File
			else
				Dir = File
			end

			local success, e = os.remove(Dir)
			if success then
				Client.Socket:send("250 Deleted FILENAME\n")
			else
				Client.Socket:send("550 Could not delete FILENAME: "..e.."\n")
			end
			return true
		elseif string.sub(line, 1, 4) == "STOR" then
			if Client.Logged then
				local File = string.sub(line, 6)
				while string.sub(File, 1, 1) == "/" do
					File = string.sub(File, 2)
				end
				while string.sub(File, string.len(File), string.len(File)) == "/" do
					File = string.sub(File, string.len(File) - 1)
				end

				if File:len() <= 0 then
					Client.Socket:send("550 No file name\n")
					return true
				end

				local Dir, Path
				if Client.Dir:len() > 0 then
					Dir = Client.Dir.."/"..File
					Path = Client.Dir
				else
					Dir = File
				end

				if not os.dirent(Path) then
					if not Path then
						Client.Socket:send("500 /"..Dir..": The system cannot find the path specified.\n")
						return true
					else
						local folders = Path:split("/")
						local NewPath = os.currentdir()
						for i = 1, #folders do
							NewPath = NewPath.."/"..folders[i]
							if not os.dirent(NewPath) then
								local success, e = os.mkdir(NewPath)
								if not success then
									Client.Socket:send("500 /"..Dir..": "..e..".\n")
									return true
								end
							end
						end
					end
				end

				if not Client.SSTART then
					Client.SSTART = os.clock()
				end

				if Client:ConnectData() then
					if not Client.SWARN then
						Client.SFILE = io.open(Dir, "w")
						Client.SWARN = true
						Client.Socket:send("150 Accepted data connection\n")
					end

					local data, e = Client.DataSocket:receive("*a")
					if data then
						if Client.SFILE then
							Client.SFILE:write(data)
						end
					elseif e == "closed" then
						Client.DataSocket = nil
						if Client.SFILE then
							Client.SFILE:close()
							Client.SFILE = nil
						end
						Client.SWARN = nil
						Client.SSTART = nil
						Client.Socket:send("226 File successfully transferred\n")
						return true
					end
				elseif os.clock() - Client.SSTART >= 2 then
					Client.SSTART = nil
					Client.SWARN = nil
					Client.Socket:send("425 Can't open data connection\n")
					return true
				end
			else
				Client.Socket:send("530 You aren't logged in\n")
				return true
			end
		elseif string.sub(line, 1, 4) == "RETR" then
			if Client.Logged then
				local File = string.sub(line, 6)
				while string.sub(File, 1, 1) == "/" do
					File = string.sub(File, 2)
				end

				if File:len() <= 0 then
					Client.Socket:send("550 No file name\n")
					return true
				end

				local Dir
				if Client.Dir:len() > 0 then
					Dir = Client.Dir.."/"..File
				else
					Dir = File
				end

				if not os.dirent(Dir) then
					Client.Socket:send("550 Can't open "..File..": No such file or directory\n")
					return true
				end

				if not Client.RSTART then
					Client.RSTART = os.clock()
				end

				if Client:ConnectData() then
					if not Client.RWARN then
						Client.RWARN = true
						Client.Socket:send("150 Accepted data connection\n")
						Client.RFILE = io.open(Dir, "r")
					end

					if Client.RFILE then
						local data = Client.RFILE:read(byte_transfer)
						if data then
							Client.DataSocket:send(data)
						else
							Client.RFILE:close()
							Client.DataSocket:close()
							Client.RSTART = nil
							Client.RWARN = nil
							Client.Socket:send("226 File successfully transferred\n")
							return true
						end
					else
						Client.RSTART = nil
						Client.RWARN = nil
						Client.DataSocket:close()
						Client.Socket:send("226 File successfully transferred\n")
						return true
					end
				elseif os.clock() - Client.RSTART >= 2 then
					Client.RSTART = nil
					Client.RWARN = nil
					Client.Socket:send("425 Can't open data connection\n")
					return true
				end
			else
				Client.Socket:send("530 You aren't logged in\n")
				return true
			end
		else
			Client.Socket:send("502 Command not implemented.\n")
			return true
		end
	end
end

function FTP.ProcessClients()
	for id, Client in pairs(FTP.Client) do
		local line, e = Client.Socket:receive("*l")
		if e == "closed" then
			local ip, port = Client.Socket:getpeername()
			print("[FTP] Client "..ip..":"..port.." disconnected")
			Client:Close()
			FTP.Client[id] = nil
		elseif line then
			table.insert(Client.CMDQUEUE, line)
		end

		for CMDID, CMD in pairs(Client.CMDQUEUE) do
			if FTP.ProcessClient(id, CMD) then
				Client.CMDQUEUE[CMDID] = nil
			end
		end
	end
end

function FTP.ProcessServer()
	if FTP.Server then
		local CurrentClient = FTP.Server:accept()
		if CurrentClient then
			local ip, port = CurrentClient:getpeername()
			print("[FTP] New Client "..ip..":"..port)

			CurrentClient:settimeout(0, "t")
			CurrentClient:settimeout(0, "b")
			CurrentClient:send("220 Hello, this is Starkkz's FTP Service\n")

			local Client = {}
			Client.Socket = CurrentClient
			Client.RenameFrom = ""
			Client.User = ""
			Client.Password = ""
			Client.Dir = ftp_dir
			Client.Logged = false
			Client.CONMODE = 0
			Client.DIP = ""
			Client.DPORT = 0
			Client.CMDQUEUE = {}

			function Client:ip()
				local ip, port = Client.Socket:getsockname()
				return ip
			end

			function Client:port()
				local ip, port = Client.Socket:getsockname()
				return port
			end

			function Client:DataIP()
				if Client.DataSocket then
					local ip, port = Client.DataSocket:getsockname()
					return ip
				end
			end

			function Client:DataPort()
				if Client.DataSocket then
					local ip, port = Client.DataSocket:getsockname()
					return port
				end
			end

			function Client:ConnectData()
				if Client.DataSocket then
					return true
				end

				if Client.CONMODE == 1 then
					if Client.PasvServer then
						local Sock = Client.PasvServer:accept()
						if Sock then
							Sock:settimeout(0, "b")
							Sock:settimeout(0, "t")
							Client.DataSocket = Sock
							return true
						end
					end
				elseif Client.CONMODE == 2 then
					local Sock, e = socket.tcp()
					if Sock then
						if Client.DIP:len() > 0 and Client.DPORT > 0 then
							Sock:settimeout(2000, "b")
							Sock:settimeout(2000, "t")
							local success, e = Sock:connect(Client.DIP, Client.DPORT)
							if success then
								Sock:settimeout(0, "b")
								Sock:settimeout(0, "t")
								Client.DataSocket = Sock
								return true
							end
						end
					end
					Sock:close()
				end
				return false
			end

			function Client:CloseData()
				if Client.DataSocket then
					Client.DataSocket:close()
					Client.DataSocket = nil
				end
			end

			function Client:ClosePasv()
				if Client.PasvServer then
					Client.PasvServer:close()
					Client.PasvServer = nil
					Client:CloseData()
				end
			end

			function Client:CloseConMode()
				if Client.CONMODE == 1 then
					Client:ClosePasv()
				elseif Client.CONMODE == 2 then
					Client:CloseData()
				end
			end

			function Client:Close()
				Client:CloseData()
				Client:ClosePasv()
				Client.Socket:close()
			end

			table.insert(FTP.Client, Client)
		end
	end
end

FTP.InitServer()

while true do
	FTP.ProcessServer()
	FTP.ProcessClients()
end
